"""
This is a part of the package for working with GT-shadows and their action on child's drawings. 
It was written jointly by Chelsea Zackey, Khanh Le, Aidan Lorenz and Vasily Dolgushev.

It contains various auxiliary functions for working with permutations and permutation groups
"""


# coding: utf-8
import time
from random import choice
"""
for a list L, choice(L) returns a random entry from L. 
It does not change L. We use it for testing
"""

from itertools import islice
"""
islice is a good tool for "slicing" generators (or sequences). 
For example, count(0) is the infinite generator which produces 0,1,2,3,...
islice(count(0),0,5) is the finite generator which produces 0,1,2,3,4
and islice(count(0),2,6) is the finite generator which produces 
2,3,4,5. 
"""

from sympy import flatten

from sympy.utilities.iterables import partitions
"""
partitions(n) is a generator of all partitions of n; partitions 
are represented as dictionaries; for example the partition (2,1,1,1) of 5
is given as {2: 1, 1: 3}.
"""

from sympy.combinatorics import Permutation as permut
"""
For example,  s = permut(7)(0,3,2)(4,6) is the permutation in S_8
with the cycle decomposition (4,6)(0,3,2)(1)(5)(7). 
Permutations always act on [0,1,2,,3,...]
So S_n acts on [0,1,..., n-1]. 

t = permut(0,1,2,3,4,5,6,7). 
The command t.size returns 8.
The command t*s returns Permutation(0, 1)(3, 6, 7)(4, 5). 
t*s is the composition s \circ t. 

"""
from sympy.combinatorics.perm_groups import PermutationGroup as PG
from sympy.combinatorics.named_groups import SymmetricGroup as SG
from sympy.combinatorics.named_groups import AlternatingGroup as AG
from sympy.combinatorics.named_groups import DihedralGroup as DG
from sympy.combinatorics.named_groups import CyclicGroup as CG
"""
For example, if t = permut(0,1,2,3,4,5,6,7) and 
s = permut(7)(0,3,2)(4,6) then G = PG((s,t)) forms the subgroup of S_8 
generated by s and t. G.order() returns 40320 (which is 8!). 
This means that s and t generate the whole S_8
t**(-1) returns the inverse of permutation t.

The command S8 = SG(8) forms the symmetric group on 8 letters. 
The command S8 == PG((s,t)) returns True

DG(n) returns the dihedral group of order 2n (as the subgroup of S_n)

CG(n) returns the cyclic group of order n generated by the cycle (0,1,...,n-1)
"""

from pickle import dump, load
#this is for storing and loading the results
"""
How to use PICKLE:
Given a Python object obJecT, we can store it in the file FILE 
by doing this:
    
dump(obJecT,open('FILE','wb'))


To unpickle the obJecT back, we do this:
    
FO = open('FILE','rb'); obJecT = load(FO)

Another option is to use the function load_now. One should 
not forget to use ' ' around the name of the file. 
"""

#Tested
def load_now(FILE):
    """
    loads (unpickles) the object from the FILE 'FileName'.
    One should not forget to use ' '.
    """
    FO = open(FILE,'rb')
    return load(FO)



from math import factorial as fact
from math import gcd


from itertools import combinations as comb
"""
comb((1,2,3,...,n), r) generates strictly increasing sequences of length r
of elements in the set {1,2,3, ..., n}.
"""

from itertools import product as prod
"""
For example, prod(L1,L2,L3) generates elements of the Cartesian 
product L1 x L2 x L3
"""

from itertools import count as count
"""
This is an infinite generator. 
For example, count(5) generates: 5,10,15,20,25, ...
"""


#tested
def display_perm(g):
    """
    g is a permutation; 
    the function prints nested tuples whose 
    entries are cycles of the permutation;
    the shift i->i+1 is incorporated
    """
    def shift(L):
        return tuple(a+1 for a in L)
    print(tuple(shift(c) for c in g.cyclic_form))


#tested
def tup2dict(p):
    """
    converts a partition p from the format of tuple to 
    the format of dictionary
    """
    out = {}; parts=list(set(p))
    for k in parts:
        out[k] = p.count(k)
    return out


#tested
def dict2tup(d):
    """
    converts the dictionary format to the tuple format;
    for instance, dict2tup({2: 1, 1: 3}) returns (1,1,1,2)
    """
    #for some reason, the keys are sorted backwards
    L=list(d.keys())
    K = sorted(L)[::-1]
    out=()
    for kk in K:
        out=out+(kk,)*d[kk]
    return out

#tested
def gener_part(n):
    """
    generates all partitions of n in the tuple format
    """
    for d in partitions(n):
        yield dict2tup(d)


#tested    
def cart_pr(L):
    """
    L is a list of iterables;
    the function generates elements of the Cartesian 
    product L[0] x L[1] x L[2]x ...
    the outputs are tuples. 
    """
    if len(L)==0:
        yield ()
    else:
        for p in cart_pr(L[:len(L)-1]):
            for e in L[len(L)-1]:
                yield p + (e,)


#tested
def lcm(a,b):
    """
    Compute the least common multiple of two numbers
    """  
    return a * b // gcd(a,b) # lcm of two numbers = product / gcd

#tested
def lcm3(a,b,c):
    """
    Compute the least common multiple of three numbers
    """ 
    return lcm(lcm(a,b),c) # lcm(a,b,c) = lcm (lcm (a,b),c)

#tested
def prm(n):
    """
    returns a random permutation in S_n
    """
    return permut.random(n)

#tested
def ran(n):
    """
    returns the tuple (0,1,2,..., n-1)
    """
    return tuple(range(n))


"""
split generates all possible 
splitting of a tuple t into parts...
For example, split((1,2,3)) generates
((1, 2, 3),)
((1, 2), (3,))
((1,), (2,), (3,))

Thus len(list(split((1,2,3,...,n)))) returns the number 
of partitions of n.
"""
#tested
def split(t):
    yield (t,)
    for m in range(len(t)-1,0,-1):
        for a in split(t[m:]):
            if m >= len(a[0]):
                yield (t[:m],) + a 
        

#tested
def toCanPerm(p):
    """
    p is a partition represented as a non-increasing tuple
    of positive integers (k1,k2,...); the function returns
    the permutation
    (0,1,...,k1-1)(k1,...k1+k2-1)...
    """
    plist =[]; pp = list(p)
    d = sum(p); i = 0
    while len(pp) != 0:
        k  = pp.pop(0)
        if k==1:
            break
        plist.append(tuple(range(i, i+k)))
        i = i+k
    return permut(plist, size=d)


#tested
def comp(s,t):
    """
    composition of two permutations in the natural order
    """
    return t*s
    
#tested
def one(n):
    """
    returns the identity element of S_n
    """
    return permut(n-1)

#tested
def is_id(g):
    """
    g is a permutation in S_n (for some n). 
    The function returns True if g is id_{S_n}. Otherwise, False  
    """
    n = g.size
    return g == one(n)

#tested
def not_id(g):
    """
    g is a permutation in S_n (for some n). 
    The function returns False if g is id_{S_n}. Otherwise, True  
    """
    return not is_id(g)


#tested    
def compAll(t):
    """
    returns the consecutive composition of all permutations in the iterable t
    """
    if len(t)==0:
        return permut(0)
    out = one(t[0].size)
    for g in t:
        out = comp(out, g)
    return out

#tested
def concat(g,h):
    """
    S_n \times S_k --> S_{n+k}
    the function concatenates two permutations
    """
    n,k = g.size,h.size 
    Lg = [i^g for i in range(n)]
    Lh = [(i^h)+n for i in range(k)]
    return permut(Lg+Lh)

#tested
def concat_tup(t,tt):
    """
    t is a tuple of permutations in S_d and tt is a tuple of 
    permutations in S_n; t and tt have the same length l
    The function returns the tuple of length l whose entries 
    obtained by "concatenating" the corresponding entries of t and tt
    """
    if len(t)!=len(tt):
        print('Sorry! The arguments of concat_tup have different lengths.')
    else:
        return tuple(concat(t[i],tt[i]) for i in range(len(t)))

#tested
def commut(g,h):
    """
    returns the group commutator g h g^{-1} h^{-1} of two permutations g,h
    """
    return compAll((g,h,g**(-1), h**(-1)))
    
#tested
def conj(g,h):
    """
    returns g h g^{-1} for two permutations g and h
    """
    return compAll((g,h,g**(-1))) 
  
#tested
def conj_tup(g,t):
    """
    return the tuple of permutations which is obtained from t by 
    conjugating every element of t by g
    """
    return tuple(conj(g,a) for a in t)

#tested
def perm_to_list(g):
    """
    g is a permutation (represented as the sympy object), say 
    g=(k_1,...,k_{r_1})(j_1,..., j_{r_2})... such that r_1>= r2 >= ...;
    the function returns the list [k_1,...,k_{r_1},j_1,..., j_{r_2},...]
    """
    L = sorted(g.full_cyclic_form, key=len)[::-1]
    return flatten(L)

#tested
def conj_by(g,gg):
    """
    g,gg are permutations in S_d of the same cycle structure;
    the function returns a permutation h in S_d such that
    conj(h,g)==gg
    """
    if g.cycle_structure!=gg.cycle_structure:
        print('The cycle structures do not match')
        return None
    d = g.size
    L=perm_to_list(g); LL=perm_to_list(gg)
    dic = {L[i]:LL[i] for i in range(d)}
    return permut([dic[i] for i in range(d)])
    

#tested
def are_conj_easy(t,tt):
    """
    t and tt are tuples (of the same length) of permutations in S_d;
    the function returns True if there exists g in S_d such that
    conj_tup(g)(t) == tt; otherwise, False. 
    """
    c=tuple(h.cycle_structure for h in t)
    cc=tuple(h.cycle_structure for h in tt)
    if c!=cc:
        return False
    d = t[0].size
    h = conj_by(tt[0],t[0])
    tup = conj_tup(h,tt)#now t[0] coincides with tup[0]
    for a in SG(d).centralizer(t[0]).elements:
        if t[1:]==conj_tup(a,tup[1:]):
            return True
    return False
    
    

#tested
def conjCl(H, G):
    """
    H and G are subgroups of S_n and H acts by conjugation on G.
    conjCl(H,G) generates representatives from orbits of this action:
    exactly one from each orbit
    """
    remP = list(G.elements)
    while len(remP)>0:
        g = remP[0]
        #print('The length of remP is ', len(remP)) # for testing
        yield g
        for h in H.elements:
            s = conj(h,g)
            if s in remP:
                remP.remove(s)
#tested
def conjCl_rk(H, G):
    """
    H and G are subgroups of S_n and H acts by conjugation on G.
    conjCl_rk(H,G) generates representatives from orbits of this action:
    exactly one from each orbit (it produces the permutaiton with the 
    smallest rank from each orbit)
    """
    remP = list(G.elements)
    while len(remP)>0:
        #print(len(remP)) # for testing
        g = remP[0]; gg = g
        for h in H.elements:
            p = conj(h,g)
            if p in remP:
                remP.remove(p)
            if p.rank()<gg.rank():
                gg = p
        yield gg

#tested
def m_units(n):
    """
    generates all residue classes m (mod n) such that 2*m+1 is 
    a unit in the ring Z/n Z 
    """ 
    for m in range(n):
        if gcd(2*m+1, n)==1:
            yield m

#tested
def trim_perms(t):
    """
    t is a tuple of permutations in S_d and G is the subgroup generated
    by elements of t. It is possible that for some indices i in {0,1...d-1}
    the G.orbit(i)=={i}. Such indices can be ignored. 
    The function returns the tuple of trimmed permutations. 
    Of course, PG(trim_perms(t)) is isomorphic to PG(t).
    """
    G = PG(t); d = t[0].size
    moved = [i for i in range(d) if len(G.orbit(i)) != 1]
    def trim(g):
        L = [moved.index(i^g) for i in moved]
        return permut(L)
    return tuple(trim(g) for g in t)
        

#tested
def double_coset_reps(H1,G, H2):
    """
    generates representatives of double cosets H1\G/H2;
    exactly one representative for each double coset
    """
    #T0=time.time() # we looked at the speed
    Gset = G.elements
    while len(Gset) != 0:
        g = Gset.pop()
        for h1 in H1.elements:
            for h2 in H2.elements:
                h1_g_h2 = comp(h1, comp(g, h2))
                if h1_g_h2 in Gset:
                    Gset.remove(h1_g_h2)
        yield g
    #print('Time for double cosets in seconds: ',(time.time()-T0)/60) #we looked at the speed

#tested (unfortunately, it works slower than double_coset_reps)
def dc_reps(H1,G, H2):
    """
    generates representatives of double coset H1gH2; 
    one representative for each coset
    """
    #T0=time.time()
    L=G.coset_transversal(H2)
    S=set(L)#we need a set version of this list; we will remove stuff from this set 
    Glist = []; k=H2.order() # k is the order of H2
    for a in L:
        Glist.extend([comp(a,h) for h in H2.elements])
    ind_list=[]#the indiced of removed elements will be dumped here
    while len(S)!=0:
        g=S.pop();
        ind_list.append(L.index(g))
        yield g
        for h in H1.elements:
            j = Glist.index(comp(h,g))//k
            if j not in ind_list:
                S.remove(L[j])
                ind_list.append(j)
    #print('Time for double cosets in seconds: ',(time.time()-T0)/60)


#used for testig generators of representatives of double cosets
def same_dc(a,b,H1,H2):
    """
    a,b are elements of G;
    the function returns True if a and b 
    belong to the same double coset; otherwise False
    """
    H2ranks={h.rank() for h in H2.elements}
    for h in H1.elements:
        el=compAll((b**(-1), h, a))
        if el.rank() in H2ranks:
            return True
    return False

#tested
def relB4(t):
    """
    t is a tuple of three permutations in S_n. 
    the function returns True if this tuple indeed represents a homomorphism from B_4 to S_n
    """
    si1, si2, si3 = t
    rel1 = (compAll((si1, si2, si1)) == compAll((si2, si1, si2)))
    rel2 = (compAll((si2, si3, si2)) == compAll((si3, si2, si3)))
    rel3 = (comp(si1,si3) == comp(si3,si1))
    return rel3 and rel1 and rel2

#tested
def restr_PB4(t):
    """
    t represents a homomorphism from B_4 to S_n.
    The function returns the tuple (x_12, x_23, x_13,...)
    which corresponds to the restriction of this homomorphism to PB_4
    """
    si1,si2,si3 = t
    x12=si1**2; x23= si2**2; x34=si3**2
    x13=compAll((si2,si1,si1,si2**(-1)))
    x24=compAll((si3,si2,si2,si3**(-1)))
    x14=compAll((si3,si2,si1,si1,si2**(-1),si3**(-1)))
    return (x12,x23,x13,x14,x24,x34)

    
#used for testing generB4
def conj_all(s):
    """
    s is a set of homomorphisms from B_4 to S_d (represented as tuples)
    The function returns the set of all conjugations of homomorphisms in 
    s by elements of S_d
    """
    for el in s:
        d=el[0].size
        break # VD: I know! It is a funny way of finding d
    return {conj_tup(g,t) for g in SG(d).elements for t in s}
    

#tested (works faster than generB4 for d>5) 
def generB4_A(d,timed):
    """
    generates tuples (si1, si2, si3) that represent homomorphisms from B4 to S_d
    """
    if timed:
        T0 = time.time()
    for t1 in split(ran(d)):
        tau1 = permut(t1); H1=SG(d).centralizer(tau1)
        for t2 in split(ran(d)):
            tau2 = permut(t2); H2=SG(d).centralizer(tau2)
            for g in double_coset_reps(H1,SG(d),H2):
                (si1,si)=(tau1,conj(g,tau2))
                if compAll((si1, si**2, si1, si**(-2))) == compAll((si**2, si1, si**(-2), si1)):
                    if si**4 == comp(si, si1)**3:
                        si2 = compAll((si, si1, si**(-1)))
                        si3 = compAll((si**2, si1, si**(-2)))
                        yield (si1,si2,si3)
    if timed:
        print((time.time()-T0)/60)


#tested
def generB4(d,timed):
    """
    generates tuples (si1, si2, si3) corresponding to homomorphisms from B4 to S_d
    Due to the relations of B_4, every element in (si1, si2, si3)
    must have the same cycle structure
    """
    if timed:
        T0 = time.time()
    for t in split(ran(d)):
        tau = permut(t); H=SG(d).centralizer(tau)
        for g in double_coset_reps(H,SG(d),H):
            (si1,si2)=(tau,conj(g,tau))
            if compAll((si1, si2, si1)) == compAll((si2, si1, si2)):
                Z = SG(d).centralizer((si1,si2))
                for gg in double_coset_reps(Z,SG(d),H):
                    si3 = conj(gg,tau)
                    if si1*si3==si3*si1:
                        if compAll((si2, si3, si2)) == compAll((si3, si2, si3)):
                            yield (si1,si2,si3)
    if timed:
        print((time.time()-T0)/60)


"""
VD: Here is how we tested generB4_A...

for d in range(2,9):
    print('The degree is ',d)
    print('I will now use generB4_A...')
    S_A = set(generB4_A(d,True))
    print('Next, I am going to use generB4...')
    S = set(generB4(d,True))
    print('Now I will test the relations of the outputs of generB4_A')
    print('Here is the result: ',{relB4(tt) for tt in S_A})
    print('    ')
    print('Let us compare the sizes of the results: ',len(S_A)==len(S))
    print('Now I will compare the outputs of generB4_A and generB4 using conj_all')
    print('Here is the result: ', conj_all(S_A)==conj_all(S))
    print('    ')
    print('    ')

"""


# used for testing generB4
def generB4slow(d):
    """
    generates all triples of permtations in S_d which satisfy the braid relations.
    Note that this may generate a lot of stuff!
    """
    for t in prod(SG(d).elements, repeat=3):
        if relB4(t):
            yield t

"""
THE CODE BELOW IS NOT USED...
"""

#still slower than double_coset_reps 
def dc_reps_new(H1,G,H2):
    #T0=time.time() # we looked at the speed
    L1=G.coset_transversal(H1)
    L2=G.coset_transversal(H2)
    S = set(prod(L1,L2))
    while len(S)!=0:
        a,b=S.pop()
        yield comp(a**(-1),b)
        prev={(a,b)}
        while True:
            new=set()
            for a1,b1 in prev:
                for g in G.generators:
                    aa,bb=comp(g,a1),comp(g,b1)
                    a2=G._coset_representative(aa,H1)
                    b2=G._coset_representative(bb,H2)
                    if (a2,b2) in S:
                        S.remove((a2,b2))
                        new.add((a2,b2))
            if len(new)==0:
                break
            prev={p for p in new}
    #print('Time for double cosets in seconds: ',time.time()-T0)


    
