"""
This package is for working with GT-shadows and their action on child's drawings.
The contributors to this package are (in the reverse alphabetic order): Chelsea Zackey,
Aidan Lorenz, Khanh Le and Vasily Dolgushev. 


Here is a link to a very useful database of Belyi maps:
http://beta.lmfdb.org/Belyi/

GT-shadows are represented as instances of the class GTsh. 
However, it is often convenient to use the instance variable. 
We denote this variable by wm (for 'word in F_2 and m'). 
wm is always a tuple with two entries: wm[0] is a word in F_2
and wm[1] is a non-negative integer. 
A GT-shadow is determined by the tuple wm and an element 
N in NFI_{PB_4}(B_4). The latter is represented by a tuple t of 
six permutations (x12, x23, x13, x14, x24, x34) in S_d.  
The latter tuple represents a group homomorphism from PB_4 to S_d 
and N is the kernel of this homomorphism. 
"""

# coding: utf-8
import time

from operator import itemgetter

from random import choice
"""
for a list L, choice(L) returns a random entry from L. 
It does not change L. We use it for testing
"""

from itertools import islice
"""
islice is a good tool for "slicing" generators (or sequences). 
For example, count(0) is the infinite generator which produces 0,1,2,3,...
islice(count(0),0,5) is the finite generator which produces 0,1,2,3,4
and islice(count(0),2,6) is the finite generator which produces 
2,3,4,5. 
"""

from sympy.utilities.iterables import partitions
"""
partitions(n) is a generator of all partitions of n; partitions 
are represented as dictionaries; for example the partition (2,1,1,1) of 5
is given as {2: 1, 1: 3}.
"""

from sympy import factorint
"""
for an integer n, factorint(n) returns the dictionary 
{p1:e1, p2:e2, ...} such that n = p1^e1 * p2^e2 *...  
"""

from sympy import primefactors
"""
for an integer n, primefactors(n) returns the sorted list of the
prime factors of n;  primefactors(-n)==  primefactors(n), 
primefactors(1)==[],  primefactors(0) ==[].     
"""

from sympy import flatten

from sympy.combinatorics import Permutation as permut
"""
For example,  s = permut(7)(0,3,2)(4,6) is the permutation in S_8
with the cycle decomposition (4,6)(0,3,2)(1)(5)(7). 
Permutations always act on [0,1,2,,3,...]
So S_n acts on [0,1,..., n-1]. 

t = permut(0,1,2,3,4,5,6,7). 
The command t.size returns 8.
The command t*s returns Permutation(0, 1)(3, 6, 7)(4, 5). 
t*s is the composition s \circ t. 

"""
from sympy.combinatorics.perm_groups import PermutationGroup as PG
from sympy.combinatorics.named_groups import SymmetricGroup as SG
from sympy.combinatorics.named_groups import AlternatingGroup as AG
from sympy.combinatorics.named_groups import DihedralGroup as DG
from sympy.combinatorics.named_groups import CyclicGroup as CG
"""
For example, if t = permut(0,1,2,3,4,5,6,7) and 
s = permut(7)(0,3,2)(4,6) then G = PG((s,t)) forms the subgroup of S_8 
generated by s and t. G.order() returns 40320 (which is 8!). 
This means that s and t generate the whole S_8
t**(-1) returns the inverse of permutation t.

The command S8 = SG(8) forms the symmetric group on 8 letters.
"""

from pickle import dump, load
#this is for storing and loading the results
"""
How to use PICKLE:
Given a Python object obJecT, we can store it in the file FILE 
by doing this:
    
dump(obJecT,open('FILE','wb'))


To unpickle the obJecT back, we do this:
    
FO = open('FILE','rb'); obJecT = load(FO)

Another option is to use the function load_now. One should 
not forget to use ' ' around the name of the file. 
"""

#Tested
def load_now(FILE):
    """
    loads (unpickles) the object from the FILE 'FileName'.
    One should not forget to use ' '.
    """
    FO = open(FILE,'rb')
    return load(FO)



from math import factorial as fact
from math import gcd
from sympy.ntheory import totient

from itertools import combinations as comb
"""
comb((1,2,3,...,n), r) generates strictly increasing sequences of length r
of elements in the set {1,2,3, ..., n}.
"""

from itertools import product as prod
"""
For example, prod(L1,L2,L3) generates elements of the Cartesian 
product L1 x L2 x L3
"""

from itertools import count as count
"""
This is an infinite generator. 
For example, count(5) generates: 5,10,15,20,25, ...
"""


from Aux import cart_pr, lcm, lcm3, m_units, prm, ran, split, comp, one, is_id, not_id
from Aux import compAll, concat, concat_tup, commut, conj, conj_tup, conjCl_rk, trim_perms

from Aux import relB4, restr_PB4, generB4, generB4_A
from Aux import dict2tup, tup2dict, are_conj_easy, toCanPerm
from Aux import double_coset_reps, gener_part, display_perm
"""
Various auxiliary functions are imported from Aux.py.

for a permutation g, display_perm(g) prints nested tuples whose 
entries are cycles of the permutation; the shift i->i+1 is incorporated;
for example, display_perm(permut(0,5)(2,4,3)) returns the nested tuple
((1, 6), (3, 5, 4))

cart_pr implements the Cartersian product, lcm (resp. lcm3) is the least common multiple
(resp. the least common multiple of three integers), m_units(n0) generates all elements 
m mod n0 such that 2*m+1 is a unit of the ring Z/n0 Z.

prm(d) returns a random permutation of degree d,
ran(n) returns the tuple (0,1,..., n-1), split(t) generates all possible splittings of a tuple t, 
comp(s,t) returns the composition of two permutations s and t with t acting first,  

one(n) returns the identity element in S_n

is_id(g) returns True if the permutation g is the identity element, otherwise False.
g.order()==1 is another way to test that g is the identity element.
 
not_id is the negation of is_id 

concat(g,h) implements the homomorphism S_n \times S_k --> S_{n+k}

concat_tup is the extension of the command concat to a pair of tuples

commut(g,h) returns the commutator of elements g and h

conj(g,h) returns g h g^{-1}

conj_tup is the extension of the command conj to the case when the second 
argument is a tuple of permutations. 

About conjCl_rk()
H and G are subgroups of S_n and H acts by conjugation on G;
for every orbit of this action, conjCl_rk(H,G) generates 
exactly one representative (more precisely, the representive with 
the smallest "rank")


Comments about trim_perms(t)...
t is a tuple of permutations in S_d and G is the subgroup generated
by elements of t. It is possible that for some indices i in {0,1...d-1}
the G.orbit(i)=={i}. Such indices can be ignored. 
The function returns the tuple of trimmed permutations. 
Of course, PG(trim_perms(t)) is isomorphic to PG(t).

relB4(t) returns True if the tuple t represents a homomorphism 
from B_4 to a symmetric group. Otherwise, it returns False.
t is a tuple of the permutations (images of sigma_1, sigma_2 and sigma_3). 
    
About restr_PB4(t)
t represents a homomorphism from B_4 to S_d.
restr_PB4(t) returns the tuple (x_12, x_23, x_13,...)
which corresponds to the restriction of this homomorphism to PB_4

About generB4(d, timed)
d is a positive integer. generB4(d, timed) generates all homomorphisms 
from B_4 to S_d up to conjugation by elements of S_d;
generB4_A(d,timed) is another version of this generator; 
generB4_A works faster than generB4 for d > 5
If generB4_A is timed then the time is displayed in minutes.


dict2tup, tup2dict; the functions dict2tup, tup2dict
allow us to reformat partitions from the dictionary format to 
the tuple format.

are_conj_easy(t,tt); 
t and tt are tuples (of the same length) of permutations in S_d;
are_conj_easy(t,tt) returns True if there exists g in S_d such that
conj_tup(g)(t) == tt; otherwise, False.


toCanPerm(p);
p is a partition represented as a non-increasing tuple of positive 
integers (k1,k2,...); toCanPerm(p) returns the permutation
(0,1,...,k1-1)(k1,...k1+k2-1)...


double_coset_reps(H1,G,H2);
the function generates representatives of double cosets H1\G/H2;
exactly one representative for each double coset
"""

#Tested            
def cenPB3(x):
    """
    x is a tuple of permutations (x12, x23, x13). 
    The function returns c : = x23 x12 x13
    """
    return compAll((x[1], x[0], x[2]))


#tested
def relPB3(t):
    """
    t is a tuple of 3 permutations: x12, x23, x13.
    The function returns True if this tuple represents a homomorphism 
    from PB_3 to the corresponding symmetric group. Otherwise, it returns False.
    """
    c = cenPB3(t)
    rel = tuple(c*g==g*c for g in t)
    return rel==(True, True, True)
    

"""
Relations for the generators of PB_4 
"""
#Tested
def rel_x12(x12,xf):
    """
    xf is a tuple of permutations (x14,x24,x34) and x12 is a permutation.
    rel_x12 returns True if the relation x_{12}^{-1} x_{i4} x_{12} = ...
    is satisfied for every i \in {1,2,3}
    """
    x14,x24,x34 = xf 
    r1 = compAll((x12**(-1),x14, x12)) == compAll((x14,x24,x14,x24**(-1),x14**(-1)))
    r2 = compAll((x12**(-1),x24, x12)) == compAll((x14,x24,x14**(-1))) 
    r3 = compAll((x12**(-1),x34, x12)) == x34
    return r1 and r2 and r3
 
#Tested
def rel_x23(x23,xf):
    """
    xf is a tuple of permutations (x14,x24,x34) and x23 is a permutation.
    rel_x23 returns True if the relation x_{23}^{-1} x_{i4} x_{23} = ...
    is satisfied for every i \in {1,2,3}
    """
    x14,x24,x34 = xf 
    r1 = compAll((x23**(-1),x14, x23)) == x14
    r2 = compAll((x23**(-1),x24, x23)) == compAll((x24,x34,x24,x34**(-1),x24**(-1))) 
    r3 = compAll((x23**(-1),x34, x23)) == compAll((x24, x34, x24**(-1)))
    return r1 and r2 and r3

#Tested
def rel_x13(x13,xf):
    """
    xf is a tuple of permutations (x14,x24,x34) and x13 is a permutation.
    rel_x13 returns True if the relation x_{13}^{-1} x_{i4} x_{13} = ...
    is satisfied for every i \in {1,2,3}
    """
    x14,x24,x34 = xf; g = commut(x14,x34)
    r1 = compAll((x13**(-1),x14, x13)) == compAll((x14,x34,x14,x34**(-1),x14**(-1)))
    r2 = compAll((x13**(-1),x24, x13)) == compAll((g,x24,g**(-1)))
    r3 = compAll((x13**(-1),x34, x13)) == compAll((x14,x34,x14**(-1)))
    return r1 and r2 and r3


#tested
def relPB4(t):
    """
    t is a tuple of six permutations in S_d;
    the function returns True if this tuple represents a homomorphism from
    PB_4 to S_d; otherwise, it returns False 
    """
    xx=t[0:3]; xf=t[3:]
    return relPB3(xx) and rel_x12(xx[0],xf) and rel_x23(xx[1],xf) and rel_x13(xx[2],xf)

#tested
def cenPB4(t):
    """
    the tuple t represents a homomorphism from PB_4 to S_d;
    the function returns the image of the generator of the center 
    of PB_4
    """
    x12, x23, x13, x14, x24, x34 = t
    return compAll((x14, x24, x34, x12, x13, x23))

#the function was used for testing relPB4
def test_relPB4(t):
    """
    the tuple t represents a homomorphism from PB_4 to S_d;
    the function returns True if cenPB4(t) commutes with every element 
    of t
    """
    c = cenPB4(t)
    out = True
    for g in t:
        out = out and (g*c==c*g)
    return out


#tested
def cap(x,y):
    """
    tuples x and y represent homomorphisms from a free group on len(x) generators
    to symmetric groups; cap(x,y) returns the tuple that represents the intersection
    of the kernels of the homomorphisms corresponding to x and y;
    we tacitly assume that len(x)==len(y).
    """
    if x==y:
        return x
    ind_x=PG(x).order(); ind_y=PG(y).order(); m=max(ind_x,ind_y)
    t = concat_tup(x,y);
    if PG(t).order()==m:
        if ind_x==m:
            return x
        else:
            return y
    return t

#tested
def Nsubgrp_less_eq(x,y):
    """
    tuples x,y represent homomorphisms from a free group on len(x) generators to 
    symmetric groups; let N_x and N_y be the kernels of these homomorphisms, respectively;
    the function returns True if N_x <= N_y; otherwise False; we tacitly assume 
    that len(x)==len(y).
    """
    if x==y:
        return True
    ind_x = PG(x).order(); ind_y = PG(y).order()
    if ind_x < ind_y:
        return False # a smaller (or =) subgroup must have a bigger (or =) index
    z = concat_tup(x,y)
    return PG(z).order()==ind_x


#Tested. It was tested using conj_tup
def sameNsubgrp(x,y):
    """
    x (resp. y) is a tuple of permutations representing a homomorphism 
    from from a free group on len(x) generators a symmetric group. 
    The function returns True if the kernels
    of the corresponding homomorphisms coincide; otherwise False;
    note that the function can be applied to homomorphisms from a finitely 
    presented group (say PB_4,PB_3, B_4, B_3) to symmetric groups;
    we tacitly assume that len(x) == len(y).
    """
    if x == y:
        return True
    indx = PG(x).order(); indy = PG(y).order()
    if indx != indy:
        return False
    z = concat_tup(x,y)
    return PG(z).order() == indx

"""
NOW coFACES:
fi... return the corresponding 
homomorphisms from PB3 to S_n
"""
#tested
def fi1_23_4(xx):
    x12, x23, x13, x14, x24, x34 = xx
    return (comp(x12, x13), comp(x24, x34), x14)

#tested
def fi123(xx):
    x12, x23, x13, x14, x24, x34 = xx
    return (x12, x23, x13)

#tested
def fi12_3_4(xx):
    x12, x23, x13, x14, x24, x34 = xx
    return (comp(x13, x23), x34, comp(x14,x24))

#tested
def fi1_2_34(xx):
    x12, x23, x13, x14, x24, x34 = xx
    return (x12, comp(x23, x24), comp(x13,x14))

#tested
def fi234(xx):
    x12, x23, x13, x14, x24, x34 = xx
    return (x23, x34, x24)
        
     
"""
#TESTS... Here is how we tested the above 5 functions fi...:
# fromPB4toS4 is a tuple of 5671 homomorphisms from PB4 to S_4
    
Test123 = [relPB3(fi123(t)) for t in fromPB4toS4]

Test234 = [relPB3(fi234(t)) for t in fromPB4toS4]

Test12_3_4=[relPB3(fi12_3_4(t)) for t in fromPB4toS4]

Test1_23_4=[relPB3(fi1_23_4(t)) for t in fromPB4toS4]

Test1_2_34=[relPB3(fi1_2_34(t)) for t in fromPB4toS4]

False in Test123+Test234+Test12_3_4+Test1_23_4+Test1_2_34 #All good!
"""

#tested
def N_PB3(t):
    """
    t represents a homomorphism from PB_4 to S_d.
    The output is a homomorphism from PB_3 to S_d whose kernel
    is the intersection of fi...^{-1}(N)
    """
    out = cap(fi123(t),fi234(t))
    out = cap(out, fi12_3_4(t))
    out = cap(out, fi1_23_4(t))
    out = cap(out, fi1_2_34(t))
    return out

#tested: we used it for testing of N_PB3 and cap
def N_PB3_1(t):
    """
    tuple t represents a homomorphism from PB_4 to S_n.
    The output is a homomorphism from PB_3 to S_n whose kernel
    is the intersection of fi...^{-1}(N);
    the output is a bit different from N_PB3: the corresponding symmetric
    group may have a bigger degree.
    """
    out = concat_tup(fi123(t),fi234(t))
    out = concat_tup(out, fi12_3_4(t))
    out = concat_tup(out, fi1_23_4(t))
    out = concat_tup(out, fi1_2_34(t))
    return out
    

""" 
Give a homomorphism from PB3 to S_d, fi...()
returns the corresponding homomorphism from PB_2 (i.e. a permutation)
"""

#tested
def fi12(x):
    x12, x23, x13 = x
    return x12

#tested
def fi12_3(x):
    x12, x23, x13 = x
    return comp(x13, x23)

#tested
def fi1_23(x):
    x12, x23, x13 = x
    return comp(x12, x13)

#tested
def fi23(x):
    x12, x23, x13 = x
    return x23

#tested
def Nord(x):
    """
    x represents a normal subgroup in PB_3;
    N_PB2 := cap... and Nord = is the index of N_PB2 in PB_2
    """
    return lcm3(fi12(x).order(), fi23(x).order(), fi12_3(x).order())

#used for testing Nord
def Nord_1(x):
    """
    Nord_1 should give the same output as Nord. It was used for testing
    """
    return lcm(fi1_23(x).order(),Nord(x))


#tested
def penta(wm, t):
    """
    t represents an element N of NFI_{PB_4}(B_4)
    and wm is a candidate for a GT-pair in GT_{pr}(N)
    The function returns True if wm satisfies the pentagon 
    relation with respect to N;
    otherwise, it returns False
    """
    x12, x23, x13, x14, x24, x34 = t
    w = wm[0]
    if w ==():
        # if f = 1 then the pentagon is satisfied
        return True
    #Now we can assume that w != ()
    x234=fi234(t)
    f234=compAll(tuple(x234[i] for i in w))
    x123=fi123(t)
    f123=compAll(tuple(x123[i] for i in w))
    x12_3_4=fi12_3_4(t)
    f12_3_4=compAll(tuple(x12_3_4[i] for i in w))
    x1_23_4=fi1_23_4(t)
    f1_23_4=compAll(tuple(x1_23_4[i] for i in w))
    x1_2_34=fi1_2_34(t)
    f1_2_34=compAll(tuple(x1_2_34[i] for i in w))
    ###
    return compAll((f234,f1_23_4,f123))==comp(f1_2_34,f12_3_4)
    

#tested
def hexa1(wm,tt):
    """
    wm is a candidate for a GT-pair and tt represents an element of NFI_{PB_3}(B_3)
    The function returns True if the first hexagon is satisfied.
    Otherwise, False.
    """
    if wm == ((),0):
        # ((),0) corresponds to the identity map PaB--> PaB
        return True
    xy = tt[:2]; x12, x23 = xy
    z = comp(x12, x23)**(-1); w,m = wm
    xz= (x12,z); yz = (x23,z) 
    fxy = compAll(tuple(xy[i] for i in w)) 
    fxz_inv = compAll(tuple(xz[i] for i in w))**(-1)
    fyz= compAll(tuple(yz[i] for i in w))
    tup = (x23**m, fxy, x12**m,fxz_inv, z**m, fyz)
    return compAll(tup).order()==1 # only the identity elem has order 1


#tested
def hexa2(wm, tt):
    """
    wm is a candidate for a GT-pair and tt represents an element of NFI_{PB_3}(B_3)
    The function returns True if the second hexagon is satisfied. 
    Otherwise, False.
    """
    xy = tt[:2]; x12, x23 = xy
    u = comp(x23, x12)**(-1); w,m = wm
    ux = (u,x12); uy = (u,x23)
    fxy_inv = compAll(tuple(xy[i] for i in w))**(-1)
    fux_inv = compAll(tuple(ux[i] for i in w))**(-1)
    fuy = compAll(tuple(uy[i] for i in w))
    tup = (fux_inv, x12**m, fxy_inv, x23**m, fuy, u**m)
    return compAll(tup).order()==1
    

#tested
def w2g(w,t):
    """
    t is a tuple of permutations in S_d of length q,
    w is a (possibly empty) word in 0,1,... q-1;
    the function returns the corresponding element on S_d;
    if q is not the length of t the function will NOT work.
    """
    d = t[0].size
    # what if w is the empty word?
    # w2g returns the identity element of S_d
    if w == ():
        return permut(d-1)
    return compAll(tuple(t[i] for i in w))


#tested
def generWF2(tt, timed=None):
    """
    tt represents a homomorphism from F_2 to S_d. 
    Let G be the permutation group generated by tt[0] and tt[1].
    For every g in G the function "yields" exactly one representative in F_2
    (as a tuple of 0s and 1s).
    Note that, if the image of this homomorphism is big, then the generator
    will work for a while.
    """
    sz = PG(tt).order()#the order of the group
    n = tt[0].size # degree of the homomorphism (n in S_n)
    if timed:
        T0 = time.time()
    G = [permut(n-1).rank()] # we will dump ranks of the new elements of G here
    W=[()]; L = [permut(n-1)] # L is the current layer of elements 
    #Note that compAll(()) = permut(0) and we need permut(n-1)
    yield () # the representative of the identity element
    while True:
        Wnew = []; Lnew = []
        for i in range(len(W)):
            g = comp(L[i],tt[0]); rg = g.rank()
            if rg not in G:
                Wnew.append(W[i] + (0,)); Lnew.append(g); G.append(rg)
            g = comp(L[i],tt[1]); rg = g.rank()
            if rg not in G:
                Wnew.append(W[i] + (1,)); Lnew.append(g); G.append(rg)
        for w in Wnew:
            yield w
        W = Wnew[:]; L = Lnew[:]
        if timed:
            print((time.time()-T0)/60,' ',len(G))#reports time for every layer in minutes
        if len(G) == sz:
            break

#tested
def word2w(p):
    """
    p is a tuple of non-negative integers;
    the function returns the tuple of zeros and 1s by 
    the rule: (k_0,k_1,k_2,...) --> (0,..,0,1,..1,0,...,0,...)
    the first k_0 entries are zeros the next k_1 entries are 1's and
    so on; p cannot be empty!
    """
    out = ()
    for i in range(len(p)):
        if i%2==0:
            out = out + ((0,)*p[i])
        else:
            out = out + ((1,)*p[i])
    return out
    


#tested
def generWComm(t, timed=None):
    """
    t is a pair of permutations in S_d.
    The function generates words in F_2 which represent distinct elements 
    of the commutator subgroup of the image of the corresponding homomorphism
    from F_2 to S_d; words in F_2 are represented as tuples of 0's and 1's.
    """
    q0, q1 = t[0].order(), t[1].order() # these are orders of images in S_d
    tupx = tuple(range(1,q0))
    tupy = tuple(range(1,q1))
    tupx0 = tuple(range(q0))
    G = PG(t); H=G.commutator(G,G)
    sz = H.order() # the size of the commutator subgroup
    ranks=[0] # we will dump ranks of images of the commutator subgroup here
    yield () # we need to generate the identity element
    if timed:
        T0 =time.time()
    for p in prod(tupx, tupy):
        w = word2w((p[0],p[1],q0-p[0],q1-p[1]))
        rk = w2g(w,t).rank()
        if rk not in ranks:
            ranks.append(rk)
            yield w
    for r in count(1):
        L = [tupx0, tupy]+([tupx,tupy]*r)
        for pp in cart_pr(L):
            px = tuple(pp[i] for i in range(0,2*r+1,2))
            py = tuple(pp[i] for i in range(1,2*r+2,2))
            lx = (-sum(px))%q0; ly = (-sum(py))%q1
            #lx and ly are the last exponets; lx cannot be zero but ly can be zero
            if lx!=0:
                if ly==0:
                    w = word2w(pp + (lx,)); rk = w2g(w,t).rank()
                    if rk not in ranks:
                        ranks.append(rk)
                        yield w
                        if timed and len(ranks)%1000==0:
                            print((time.time()-T0)/60, ' I found ',len(ranks),' words.')
                else:
                    w = word2w(pp + (lx,ly)); rk = w2g(w,t).rank()
                    if rk not in ranks:
                        ranks.append(rk)
                        yield w
                        if timed and len(ranks)%1000==0:
                            print((time.time()-T0)/60, ' I found ',len(ranks),' words.')
            if len(ranks)==sz:
                break
        if len(ranks)==sz:
            break



#tested
def gener_GT_pr(t):
    """
    t represents an element N of NFI_{PB_4}(B_4);
    the function generates all friendly GT-pairs with the target N 
    and f in F_2
    """
    tPB3 = N_PB3(t); N0 = Nord(tPB3)
    for w in generWF2(tPB3[:2], timed=True):
        if penta((w,0),t):
            for m in m_units(N0):
                if hexa1((w,m),tPB3) and hexa2((w,m),tPB3):
                    yield (w,m)



#tested
def sourcePB3(wm, tt):
    """
    the tuple wm represents a GT-pair and 
    the tuple tt represents an element of NFI_{PB_3}(B_3)
    The function returns a tuple which represents 
    the element of NFI_{PB_3}(B_3) which is the "source" of 
    the corresponding GT-pair.
    """
    if wm ==((),0):
        # when wm is the identity
        return tt
    x12, x23, x13 = tt; w,m = wm; d = x12.size
    if w == ():
        f=one(d); g=one(d)
    else:
        f = compAll(tuple((x12,x23)[i] for i in w))
        g = compAll(tuple((x12,x13)[i] for i in w))
    y12 = x12**(2*m+1)
    y23 = compAll((f**(-1),x23**(2*m+1),f))
    y13 = compAll((x12**(-m), g**(-1),x13**(2*m+1),g, x12**m))
    return (y12, y23, y13)
    

#tested
def sourcePB4(wm, tt):
    """
    wm  represents a GT-pair and tt represents an elements of NFI_{PB_4}(B_4)
    The function returns a tuple which represents the elements of
    NFI_{PB_4}(B_4) which is the "source" of sh.    
    """
    if wm ==((),0):
        # when wm is the identity
        return tt
    x12, x23, x13, x14, x24, x34 = tt; w,m = wm; d=x12.size
    x1_23 = comp(x12,x13)
    if w ==():
        f = one(d); g = one(d); g_shft = one(d)
        f12_3_4 = one(d); f1_23_4=one(d); h = one(d)
    else:
        f = compAll(tuple((x12,x23)[i] for i in w))
        g = compAll(tuple((x12,x13)[i] for i in w))
        g_shft = compAll(tuple((x23,x24)[i] for i in w))
        z = fi12_3_4(tt)
        f12_3_4 = compAll(tuple(z[i] for i in w))
        u = fi1_23_4(tt)
        f1_23_4 = compAll(tuple(u[i] for i in w))
        h = compAll(tuple((x1_23,x14)[i] for i in w)) # this is f_{xy}(x12 x13, x14)
    y12 = x12**(2*m+1)
    y23 = compAll((f**(-1),x23**(2*m+1),f))
    y13 = compAll((x12**(-m), g**(-1),x13**(2*m+1),g, x12**m))
    y14 = compAll((f**(-1), x1_23**(-m), h**(-1),x14**(2*m+1),h, x1_23**m, f))
    y24 = compAll((f**(-1), f1_23_4**(-1), x23**(-m), g_shft**(-1),x24**(2*m+1),g_shft, x23**m, f1_23_4, f))
    y34 = compAll((f12_3_4**(-1), x34**(2*m+1), f12_3_4))
    return (y12, y23, y13, y14, y24, y34)


#tested
def gener_GT_sh(t):
    """
    t represents an element N in NFI_{PB_4}(B_4).
    The function generates all GT-shadows with f in F_2
    corresponding to N. These GT-shadows are represented 
    as pairs (word, integer).
    """
    tPB3 = N_PB3(t); N0 = Nord(tPB3)
    sz = PG(tPB3).order()
    for w in generWF2(tPB3[:2], timed=True):
        if penta((w,0),t):
            for m in range(N0):
                if gcd(2*m+1,N0)==1:
                    if hexa1((w,m),tPB3) and hexa2((w,m),tPB3):
                        src = sourcePB3((w,m),tPB3)
                        if PG(src).order()==sz:
                            yield (w,m)
    

#tested
def gener_GT_charm(t):
    """
    tuple t represents an element N in NFI_{PB_4}(B_4).
    The function generates all charming GT-shadows
    corresponding to N;  the function uses the generator generWComm
    """
    tPB3 = N_PB3(t); N0 = Nord(tPB3)
    sz = PG(tPB3).order()
    for w in generWComm(tPB3[:2],True):
        if penta((w,0),t):
            for m in range(N0):
                if gcd(2*m+1,N0)==1:
                    if hexa1((w,m),tPB3) and hexa2((w,m),tPB3):
                        src = sourcePB3((w,m),tPB3)
                        if PG(src).order()==sz:
                            yield (w,m)
    



#tested
def gener_GT_penta(t):
    """
    t represents an element N in NFI_{PB_4}(B_4).
    The function generates all elements in the commutator subgroup of F_2/N_{F_2} 
    (represented by words in [F_2,F_2]) which satisfy the pentagon relation
    with respect to N. Note that the generator generWComm is used. 
    """
    tPB3 = N_PB3(t)
    #sz = PG(tPB3).order()
    #szF2 = PG(tPB3[:2]).order()
    for w in generWComm(tPB3[:2],True):
        if penta((w,0),t):
            yield w


    
"""
t is a tuple which represents an element 
of NFI_{PB_4}(B_4). An instance of the class Equiv represents 
the corresponding compatible equivalence relation on PaB(<= 4) 

self.PB3 represents N_{PB3}

self.N0 is N_{ord}

self.xy represents the corresponding homomorphism from F_2

self.d4 (resp. self.d3) is the "degree" of permutations in t (resp. N_PB3(t))
"""
class Equiv:
    def __init__(self, t):
        self.PB4=t; self.PB3 = N_PB3(t)
        self.N0 = Nord(self.PB3)
        self.xy = self.PB3[:2]
        self.d4 = t[0].size; self.d3 = self.PB3[0].size
    #tested    
    def __eq__(self, other):
        return sameNsubgrp(self.PB4, other.PB4)
    #tested
    def ind4(self):
        """
        returns the index of N in PB_4
        """
        return PG(self.PB4).order()
    #tested
    def ind3(self):
        """
        returns the index of N in PB_3
        """
        return PG(self.PB3).order()
    #tested
    def indF2(self):
        """
        returns the index of N_{PB_3} cap F_2 in F_2
        """
        return PG(self.xy).order()
    #tested
    def commF2(self):
        """
        returns the commutator subgroup of F_2/N_{F_2}
        """
        G=PG(self.xy)
        return G.commutator(G,G)
    
    #tested
    def ord_x(self):
        """
        returns the order of the coset of x12 in PB3/N_{PB_3} 
        """
        x = self.PB3[0]
        return x.order()
    #tested
    def ord_y(self):
        """
        returns the order of the coset of x23 in PB3/N_{PB_3} 
        """
        y = self.PB3[1]
        return y.order()
    
    #tested
    def relPB4(self):
        """
        returns True if t = self.PB4 represents a homomorphism from PB4
        """
        return relPB4(self.PB4)
    #tested
    def finer(self, E):
        """
        returns True if self is a finer equivalence relation than E.
        Otherwise False.
        """
        return Nsubgrp_less_eq(self.PB4,E.PB4)
    #tested
    def cap(self,E):
        """
        returns the intersection of self with E
        """
        tt = cap(self.PB4, E.PB4)
        return Equiv(tt)
    
    #tested
    def subord(self,c):
        """
        returns True if the permutation pair c is subordinate to the equivalence
        relation self; i.e., F_2 \cap N_{PB_3} is a subgroup of the kernel of 
        the homomorphism from F_2 corresponding to c; otherwise the command returns False.  
        """
        t = self.xy #ker of t is F_2 \cap N_{PB3}
        ind_t = PG(t).order(); ind_c = PG(c).order()
        # the index of the subgroup divides the order of the group.
        if ind_t%ind_c != 0:
            return False
        cap = (concat(t[0],c[0]),concat(t[1],c[1])) # represents the intersection of kernels
        return PG(cap).order()==ind_t

        
#tested
def fishingE(L,d,timed):
    """
    L is a list of equivalence relations, d is a positive integer
    The procedure looks through (conjugacy classes of) homomorphisms 
    form B_4 to S_d and intersects the corresponding kernels with PB_4;
    If the resulting equivalence relation has PB4/N >= 6 and 
    it is not in L then the procedure appends it to L
    """
    for tB in generB4_A(d,timed):
        tt = restr_PB4(tB)
        if PG(tt).order()>=6:
            E=Equiv(tt)
            if E not in L:
                print('  I found an interesting equivalence relation')
                print('  the index of N_F2 is ', E.indF2())
                L.append(E)

        
#tested
def fishingE_nonAb(L,d,timed):
    """
    This is a variation of fishingE;
    the procedure also tests whether F_2/N_{F_2} is non-Abelian. 
    The procedure appends the equivalence relation to L 
    only if F_2/N_{F_2} is non-Abelian
    """
    for tB in generB4_A(d,timed):
        tt = restr_PB4(tB)
        if PG(tt).order()>=6:
            E=Equiv(tt); G=PG(E.xy)
            if G.is_abelian:
                pass
            else:
                if E not in L:
                    print('  I found an interesting equivalence relation')
                    print('  the index of N_F2 is ', E.indF2())
                    L.append(E)


"""
We used fishingE for d = 4, 5, 6, 7, 8.
A computer worked more than a day for d=9. 
I did not let the computer finish the case d=9
(We could "fish" more elements of NFI_{PB_4}(B_4) by 
looking though homomorphisms B_4 --> S_9). 
"""

        
"""
wm is a tuple (w,m) where w in a word in F_2 and m is a non-negative integer
t is the tuple which represents the corresponding elements of NFI_{PB_4}(B_4)

self.tar is the compatible equivalence relation corresponding 
to the target of self

For example, in order to get the tuple which represents N in NFI_{PB4}(B_4),
we need to write self.tar.PB4

self.g is the image of w (i.e. f) in F_2/N_{PB_3} cap F_2

"""

class GTsh:
    def __init__(self, wm,t):
        self.wm=wm; self.w = wm[0]; self.m=wm[1]
        self.tar = Equiv(t)
        self.cc_ch = (2*wm[1]+1)%self.tar.N0
        if wm[0]==():
            self.g = one(self.tar.d3)
        else:
            self.g = compAll(tuple(self.tar.xy[i] for i in wm[0]))
    #tested
    def __eq__(self, other):
        if self.m != other.m or self.tar != other.tar:
            return False
        if other.wm[0]==():
            g = one(self.tar.d3)
        else:
            g = compAll(tuple(self.tar.xy[i] for i in other.wm[0]))
        return g == self.g
    
    #tested
    def is_GTpr(self):
        """
        returns True if the instance is a friendly GT-pair. Otherwise, False.
        """
        return gcd(2*self.m+1,self.tar.N0)==1 and penta(self.wm,self.tar.PB4) and hexa1(self.wm,self.tar.PB3) and hexa2(self.wm,self.tar.PB3)
    
    #tested
    def src(self):
        """
        returns the source of the GT-shadow viewed as a morphism to self.PB4
        """
        return Equiv(sourcePB4(self.wm,self.tar.PB4))
    
    #tested
    def is_GTsh(self):
        """
        returns True if sh is a GT-shadow in GT(N) where N is the kernel
        of the homomorphism corresponding to t.
        """
        Et = self.tar; Es = self.src()
        return self.is_GTpr() and Es.ind3() == Et.ind3()
    
    #tested
    def is_charm(self):
        """
        returns True if self is a charming GT-shadow
        """
        Et = self.tar
        G=PG(Et.xy); H=G.commutator(G,G)
        szF2 = PG(sourcePB3(self.wm,Et.PB3)[:2]).order()
        return self.is_GTsh() and self.g in H.elements and Et.indF2()==szF2
       
    #tested
    def settled(self):
        """
        returns True if the GT-shadow self is settled, i.e. the source 
        coincides with the target
        """
        return self.src()==self.tar
    #tested
    def proj(self,E):
        """
        E represents a compatible equivalence relation on PaB(<=4)
        The method returns the image of the shadow 
        self in GT(N_E) where N_E is the normal subgroup of PB_4
        corresponding to E. We assume that E is coarser than self.tar
        """
        m1 = self.m % E.N0 # we need to consider m modulo the N_ord for E
        return GTsh((self.w,m1),E.PB4)
    #tested
    def survives(self,E):
        """
        E is a finer equivalence relation.
        The method returns True if there exists a GT-shadow in GT(E)
        whose projection is self. Otherwise False.
        NOTE: It may work slowly...
        """
        for wm in gener_GT_charm(E.PB4):
            if GTsh(wm, E.PB4).proj(self.tar) == self:
                return True
        return False
    #tested    
    def compose(self,other):
        """
        computes the compostion "self circ other" provided 
        the GT-shadows are composable. The output is an instance of the 
        class GTsh
        """
        if self.src() != other.tar:
            print('The GT-shadows cannot be composed')
        else:
            N0 = other.tar.N0; d_x=self.tar.ord_x() 
            d_y=self.tar.ord_y(); w1 = self.w; w2 = other.w
            m1 = self.m; m2 = other.m
            w1_inv = () # we need to compute the "inverse" of w1
            for s in w1[::-1]:
                if s == 0:
                    w1_inv = w1_inv + (0,)*(d_x-1)
                else:
                    w1_inv = w1_inv + (1,)*(d_y-1)
            w2_new = ()
            for s in w2:
                if s==0:
                    w2_new = w2_new + (0,)*(2*m1+1)
                else:
                    w2_new = w2_new + w1_inv + (1,)*(2*m1+1) + w1
            w = w1+w2_new
            m = (2*m1*m2+m1+m2)%N0 
            return GTsh((w,m), self.tar.PB4)
    #tested
    def inv(self):
        """
        returns the inverse of self (assuming that self is charming);
        this is a "bulldozer way" of looking for the inverse;
        so it may work very slowly
        """
        E = self.src(); tt = E.PB4
        I = GTsh( ((),0) ,self.tar.PB4)# the identity element with same target as self
        if self == I:
            return self
        for wm in gener_GT_charm(tt):
            P = GTsh(wm,tt)
            if P.src() == self.tar:
                if self.compose(P) == I:
                    return P
    #tested            
    def act(self,c):
        """
        c is a child's drawing in the tuple format and it is subordinate to self.tar;
        the method returns the result of the action of self on c and output is in the 
        tuple format
        """
        m = self.m
        if self.w==():
            return (c[0]**(2*m+1), c[1]**(2*m+1))
        x,y = c; 
        xnew = x**(2*m+1)
        g = compAll(tuple(c[i] for i in self.w))
        ynew = compAll((g**(-1), y**(2*m+1), g))
        return (xnew, ynew)

#tested
def test_isolated(L):
    """
    L is a complete list of GT-shadows 
    (represented as instances of the class GTsh)
    The function returns True if every GT-shadow in L is settled. 
    """
    for P in L:
        if not P.settled():
            return False
    return True
    

#tested
def shadows2perm_group(L,timed=None):
    """
    L is a complete list of (charming) GT shadows 
    for a given ISOLATED element N of NFI_{PB_4}(B_4);
    the function returns the corresponding permutation group
    as a subgroup of S_d where d is the length of L;
    note that d is also the order of this permutation group
    """
    if timed:
        t0=time.time()
    d=len(L); perm_list=[]
    LL =[T for T in L if T.wm != ((), 0)]
    #we know that, if T.wm == ((),0) then T gives the identity element
    for T in LL:
        aux = [L.index(T.compose(TT)) for TT in L]
        perm_list.append(permut(aux))
        G=PG(perm_list)
        print(G.order())#to see how it is doing
        if G.order()==d:
            if timed:
                print('time in minutes is ',(time.time()-t0)/60) 
            return G

#tested
def shadows2perm_group_choice(L,timed=None):
    """
    This is a version of shadows2perm_group in which the function
    draws non-identity elements from L randomly; 
    we assume that L[0] is the identity element
    """
    if timed:
        t0=time.time()
    d=len(L); perm_list=[]
    ind = list(range(1,d))#indices of non-identity elements
    #we know that, if T.wm == ((),0) then T gives the identity element
    for a in range(d-1):
        i=choice(ind); ind.remove(i); T=L[i]
        aux = [L.index(T.compose(TT)) for TT in L]
        perm_list.append(permut(aux))
        G=PG(perm_list)
        print(G.order())#to see how it is doing
        if G.order()==d:
            if timed:
                print('time in minutes is ',(time.time()-t0)/60) 
            return G

#tested
def shadow2perm(T,L):
    """
    L is a complete list of (charming) GT-shadows
    for a fixed (isolated) element of NFI_{PB_4}(B_4)
    and T is an element of this list; the function returns
    the permutation in S_{len(L)} corresponding to the
    left action of T on L; we assume that L[0] is
    the identity element.
    """
    d = len(L)
    if T.wm==((), 0):
        return one(d)
    aux = [L.index(T)]+[L.index(T.compose(TT)) for TT in L[1:]]
    return permut(aux)


"""
Looking for charming GT-shadows that are fake...
"""
#tested
def charm_fake_search(E,EE,num):
    """
    EE is a finer compatible equivalence relations than E
    and num is the total number of charming GT-shadows
    whose target is E; the function returns True 
    if all charming GT-shadows with target E survive into EE;
    Otherwise, False.
    The function searches for charming GT-shadows 
    with the target EE and projects them to E 
    """
    if EE.finer(E):
        pass
    else:
        print('Sorry! Your second argument is not finer than the first one.')
        return None
    Test =[]
    for wm in gener_GT_charm(EE.PB4):
        m=wm[1]%E.N0
        T=GTsh((wm[0],m),E.PB4) #another way to compute the projection
        if T not in Test:
            Test.append(T)
            #uncomment the lines below if you want to monitor the process
            #print("The current number of charming")
            #print("GT-shadows which survive into ... is ",len(Test)) #to see how we are doing...
        if len(Test)==num:
            return True
    return False

#tested
def charm_fake4isolated(E,EE,num):
    """
    instances E, EE of the class GTsh represent isolated 
    elements of  NFI_{PB4}(B_4); the equivalence relation 
    represented by EE is finer than the one represented by E;
    num is the total number of charming GT-shadows
    whose target is E; the function returns True 
    if all charming GT-shadows with target E survive into EE;
    otherwise, the function returns False; 
    the function generates charming GT-shadows 
    with the target EE and "projects them" to E; 
    once it finds >num//2 distinct elements of GT(E), it returns True
    """
    if EE.finer(E):
        pass
    else:
        print('Sorry! Your second argument is not finer than the first one.')
        return None
    Test =[]
    for wm in gener_GT_charm(EE.PB4):
        m=wm[1]%E.N0
        T=GTsh((wm[0],m),E.PB4) #another way to compute the projection
        if T not in Test:
            Test.append(T)
            #uncomment the lines below if you want to monitor the process
            #print("The current number of charming")
            #print("GT-shadows which survive into ... is ",len(Test)) #to see how we are doing...
        if len(Test) > num//2:
            return True
    return False
    
"""
THE CODE BELOW IS FOR TESTING VERSIONS OF FURUSHO PROPERTY...
"""
    
#tested
def onto(w,E):
    """
    w represents a word in F_2 and E is a compatible equiv relation;
    the function returns True if there exists m such that 2m+1 is a unit 
    of Z/N_{ord} Z and the homomorphism F_3 --> PB_3/N_{PB_3} corresponding to
    (w,m) is onto
    """
    d3=E.ind3()
    for m in m_units(E.N0):
        tt = sourcePB3((w,m), E.PB3)
        if PG(tt).order()==d3:
            return True
    return False

     

#tested
def furusho(w,E):
    """
    E is a compatible equivalence relation on PaB(<=4)
    and w is a word in F_2 which satisfies the pentagon
    relation (mod E); the function returns True if the 
    set of m's for which both hexagons are satisfied is non-empty.
    """
    n0 = E.N0; tPB3 = E.PB3
    for m in m_units(n0):
        if hexa1((w,m),tPB3) and hexa2((w,m),tPB3):
            return True
    return False
        
"""
Testing the Furusho phenomenon in the set-up of
all (practical) GT-shadows. 
"""
#tested
def furusho_test(E):
    """
    E is a compatible equivalence relation on PaB(<=4);
    the function generates tuples of length 2; the first entry
    of each tuple is an element of  F_2/N_{F_2} (represented 
    as words in F_2) that satisfies the pentagon relation and 
    the second entry is the output of furusho( ,E), i.e. True or False
    """
    for w in generWF2(E.xy,True):
        if penta((w,0),E.PB4):
            yield (w,furusho(w,E))

#tested
def furusho_test_comm(E):
    """
    E is a compatible equivalence relation on PaB(<=4);
    the function generates tuples of length 2; the first entry
    of each tuple is an element of  [F_2/N_{F_2}, F_2/N_{F_2}] 
    (represented as words in F_2) which satisfy the pentagon relation and 
    the second entry is the output of furusho( ,E), i.e. True or False
    """
    for w in generWComm(E.xy,True):
        if penta((w,0),E.PB4):
            yield (w,furusho(w,E))
    
"""
Below are faster functions for testing Furusho
properties; they return less information... 
"""
    
#tested
def furusho_test1(E):
    """
    E is a compatible equivalence relation on PaB(<=4);
    the function returns False if at least one element 
    of F_2/N_{F_2} does not satisfy the Furusho property;
    if all elements of F_2/N_{F_2} satisfy the Furusho 
    property then the function returns True
    """
    for w in generWF2(E.xy,True):
        if penta((w,0),E.PB4) and furusho(w,E)==False:
            return False
    return True

#tested
def furusho_test_comm1(E):
    """
    E is a compatible equivalence relation on PaB(<=4);
    if the function finds at least one element in 
    [F_2/N_{F_2}, F_2/N_{F_2}] for which the Furuhso 
    phenomenon does not hold, then it returns False;
    if all elements of [F_2/N_{F_2}, F_2/N_{F_2}] satisfy 
    the Furusho property, the function returns True;
    it make take a while to get True...
    """
    for w in generWComm(E.xy,True):
        if penta((w,0),E.PB4) and furusho(w,E)==False:
            return False
    return True


#tested
def grpd_GT(t):
    """
    t is represents an element of NFI_{PB_4}(B_4).
    The function returns a list with two entries; the first
    entry is the list of objects of the corresponding 
    groupoid (saved as instanced of the class Equiv) and the second is 
    the list of morphisms, i.e. charming GT-shadows 
    (saved as instances of the class GTsh)
    """
    E = Equiv(t); Mor = []; Ob = [E]; prevOb =[E] 
    while True:
        newOb = [] # we will dump new objects into this list
        for E in prevOb:
            for wm in gener_GT_charm(E.PB4):
                P = GTsh(wm,E.PB4); EE = P.src()
                Mor.append(P)
                if EE not in Ob:
                    Ob.append(EE); newOb.append(EE)
        if len(newOb)==0:
            break
        prevOb = newOb[:]
    return [Ob,Mor]
            
            
"""
BEWOW ARE THE FUNCTIONS FOR WORKING WITH permutation triples/pairs and 
child's drawings

A permutation triple of degree d is represented by a tuple (g,gg) of permutations in S_d 
such that the group generated by g, gg acts transitively on 
{0,1,..., d-1}.

Two permutation triples c, cc (of the same degree d) are equivalent if 
there exists h in S_d such that conj_tup(h,c)==cc. 

An equivalence class of a permutation triple is called 
a child's drawing.
"""

"""
Instances of the class Dessin represent child's drawings. 
The instance variable pr is the tuple of two permutations in S_d

self.d returns the degree of the perm. triple;

self.full returns the full permutation triple;

self.passport returns the passport of self (each partition 
is given in the tuple format).
"""

class Dessin:
    def __init__(self, pr):
        self.pr=pr; self.d=pr[0].size
        self.full=pr + (compAll(pr)**(-1),)
        self.passport = tuple(dict2tup(g.cycle_structure) for g in self.full)
    #tested  
    def __eq__(self, other):
        if self.d!=other.d:
            return False
        if other.pr==self.pr:
            return True
        return are_conj_easy(self.pr, other.pr)
    #tested
    def monG(self):
        """
        returns the monodromy group of self
        """
        return PG(self.pr)
    #tested
    def is_transitive(self):
        """
        returns True if the monodromy group of self indeed acts
        transitively on {0,1,...,d-1}; otherwise False
        """
        return PG(self.pr).is_transitive()
    #tested
    def genus(self):
        """
        returns the genus of self; see formula 1.2 
        on page 23 of Graphs on Surfaces (Lando and Zvonkine)
        """
        d = self.d; pt=self.passport
        v = [(d-len(t)) for t in pt]
        return 1 - d + sum(v)//2
    #tested
    def is_Galois(self):
        """
        returns True if self corresponds to a Galois covering of P^1-{0,1,\infty};
        otherwise, False
        """
        return PG(self.pr).order()==self.d


#tested
def dessin_star(d):
    """
    returns the tuple (permut(0,1,.. d-1), id_d)
    """
    return (permut((ran(d),)), permut(d-1))

#tested
def dessin_path(d):
    """
    for a positive integer d, dessin_path(d) returns the path 
    dessin of degree d in the tuple format;
    """
    if d==1:
        return (one(1), one(1))
    if d==2:
        return (one(2), permut(0,1))
    if d%2==1:
        L1=[(2*k-1,2*k) for k in range(1,1+((d-1)//2))]
        L2=[(2*k,2*k+1) for k in range(1+((d-3)//2))]+[(d-1,)]
        return (permut(L1), permut(L2))
    L1=[(2*k-1,2*k) for k in range(1,1+((d-2)//2))]+[(d-1,)]
    L2=[(2*k,2*k+1) for k in range(1+((d-2)//2))]
    return (permut(L1), permut(L2))
    

#tested
def gener_dessin_pt(tup):
    """
    tup is a tuple (c1,c2,c3) of partitions of d (represented as tuples);
    the function generates child's drawings whose passport is tup 
    """
    d = sum(tup[0])
    g1 = toCanPerm(tup[0])
    g2 = toCanPerm(tup[1])
    c3 = tup2dict(tup[2])
    Z1 = SG(d).centralizer(g1)
    Z2 = SG(d).centralizer(g2)
    for g in double_coset_reps(Z1, SG(d), Z2):
        h = conj(g,g2)
        f = comp(g1,h)**(-1)
        if PG(g1, h).is_transitive() and f.cycle_structure == c3:
            yield (g1, h)
            

#tested
def find_dessin_pt_random(tup,n, L):
    """
    tup is a tuple (c1,c2,c3) of partitions of d (represented as tuples),
    n is a positive integer and L is a list of distinct child's drawings
    whose passport == tup;
    the procedure runs a loop n times; in each iteration it tries 
    to generate a random child's drawing with the passport tup;
    if it finds a child's drawing that does not belong to L, then it 
    appends it to L
    """
    d = sum(tup[0])
    a=tup[1][0]; b=tup[1][0]+tup[1][1]
    for i in range(n):
        seq=list(prm(d))
        g=permut([seq[:a],seq[a:b],seq[b:d]])
        c=(toCanPerm(tup[0]),g)
        if PG(c).is_transitive():
            D=Dessin(c)
            if D.passport==tup and D not in L:
                L.append(D)



#tested
def gener_dessin_slow(d):
    """
    d is an integer > 1;
    this is a generator of all child's drawings of degree d in the tuple format;
    it is based on gener_dessin_pt and it is slower than gener_dessin below
    """
    p_list=[dict2tup(p) for p in partitions(d)]
    part111=tuple(1 for i in range(d))
    for pt in prod(p_list,p_list,p_list):
        if pt != (part111,part111,part111):
            for c in gener_dessin_pt(pt):
                yield c

#tested
def gener_dessin(d):
    """
    generates all child's drawing of degree d in the tuple format
    """
    Sd = SG(d) # Symmetric Group of degree d
    for x in split(ran(d)):
        g = permut(x) # Pick the first permutation based on partition of d
        for gg in conjCl_rk(Sd.centralizer(g),Sd):
            if PG(g,gg).is_transitive():  # if the group generated by g and gg is transitive then yield (g,gg)
                yield (g,gg)
        
#tested
def all_dessin(d):
    """
    returns the list of all child's drawings of degree d; 
    each entry is in the tuple format;
    the function also prints the elapsed time (in minutes)
    """
    t0=time.time()
    L=list(gener_dessin(d))
    print((time.time()-t0)/60)
    return L
    
#tested
def all_dessin_slow(d):
    """
    returns the list of all child's drawings of degree d; 
    each entry is in the tuple format;
    the function also prints the elapsed time (in minutes);
    this function is slower than all_dessin above
    """
    t0=time.time()
    L=list(gener_dessin_slow(d))
    print((time.time()-t0)/60)
    return L
    

#tested
def rand_dessin(d):
    """
    returns a somewhat random child's drawing of degree d
    """
    while True:
        g = prm(d); gg = prm(d)
        if PG(g,gg).is_transitive():
            return (g,gg)

#tested
def orbit(c,L):
    """
    c is a tuple that represents a child's drawing subordinate to N 
    and L is a list of GT-shadows (in the object format) whose target is N;
    the command orbit(c,L) returns the corresponding orbit of [c]; the output is
    a list and each entry of the list is an instance of the class Dessin
    """
    orb=[Dessin(c)]
    for T in L:
        D=Dessin(T.act(c))
        if D not in orb:
            orb.append(D)
    return orb


#tested
def srch_dessin(E,d):
    """
    E is an instance of the class Equiv and d is a positive integer;
    it is a generator of all (if any) permutation pairs of degree d
    (in the canonical form) whose child's drawings are subordinate to E
    """
    Sd = SG(d)
    xy = E.xy; sz = PG(xy).order()
    for t in split(ran(d)):
        g = permut(t)
        if sz%g.order()==0:
            for gg in conjCl_rk(Sd.centralizer(g),Sd):
                h = g*gg
                if sz%gg.order()==0 and sz%h.order()==0:
                    if PG(g,gg).is_transitive():
                        c = (g,gg)
                        if sz%PG(c).order()==0:
                            tup = cap(xy,c)
                            if PG(tup).order()==sz:
                                yield c
                            

#tested
def conjBySig1(x):
    """
    x represents represents a normal subgroup N of PB4 (really, it represents
    a homomorphims PB4 --> Sn where x is a 6-tuple comprised of the images
    of the generators of PB4 and the normal subgroup N is the kernel of this
    homomorphism). The function returns the normal subgroup
    \sigma_{1}^{-1} N \sigma_{1}
    """
    x12, x23, x13, x14, x24, x34 = x
    return (x12, compAll((x12,x13,x12**(-1))), x23, x24, compAll((x12,x14,x12**(-1))), x34)

#tested
def conjBySig2(x):
    """
    x represents represents a normal subgroup N of PB4 (really, it represents
    a homomorphims PB4 --> Sn where x is a 6-tuple comprised of the images
    of the generators of PB4 and the normal subgroup N is the kernel of this
    homomorphism). The function returns the normal subgroup
    \sigma_{2}^{-1} N \sigma_{2}
    """
    x12, x23, x13, x14, x24, x34 = x
    return (x13, x23, compAll((x23,x12,x23**(-1))), x14, x34, compAll((x23,x24,x23**(-1))))

#tested
def conjBySig3(x):
    """
    x represents represents a normal subgroup N of PB4 (really, it represents
    a homomorphims PB4 --> Sn where x is a 6-tuple comprised of the images
    of the generators of PB4 and the normal subgroup N is the kernel of this
    homomorphism). The function returns the normal subgroup
    \sigma_{3}^{-1} N \sigma_{3}
    """
    x12, x23, x13, x14, x24, x34 = x
    return (x12, x24, x14, compAll((x34,x13,x34**(-1))), compAll((x34,x23,x34**(-1))), x34)

#tested        
def isNormB4(x):
    """
    x represents represents a normal subgroup N of PB4 (really, it represents
    a homomorphims PB4 --> Sn where x is a 6-tuple comprised of the images
    of the generators of PB4 and the normal subgroup N is the kernel of this
    homomorphism); the function returns True if N is normal in B4; 
    otherwise, the function returns False
    """
    return sameNsubgrp(x, conjBySig1(x)) and sameNsubgrp(x, conjBySig2(x)) and sameNsubgrp(x, conjBySig3(x))

#tested
def dessin2PB4(c):
    """
    c is a pair of permutations in S_d
    The function returns a homomorphism from PB_4 to S_d
    such that \vf^{-1}_{123}(ker)\cap F_2 = ker(F_2 --> S_d)
    """
    x12, x23 = c; x13 = comp(x12**(-1),x23**(-1))
    d = x12.size
    return (x12, x23, x13, one(d), one(d), one(d))
    

#tested
def B4_inv(tt):
    """
    tt represents a normal subgroup (of finite index) N of PB_4;
    the function returns an element K in NFI_{PB_4}(B_4)
    such that K \le N
    """
    out = tt; q=0
    while True:
        F1=False; F2=False; F3=False
        q = q+1; print(q)
        #Sig1
        new1 = conjBySig1(out)
        if not sameNsubgrp(out, new1):
            out = concat_tup(out,new1)
        else:
            F1=True
        #now Sig2
        new2 = conjBySig2(out)
        if not sameNsubgrp(out, new2):
            out = concat_tup(out,new2)
        else:
            F2=True
        #now Sig3
        new3 = conjBySig3(out)
        if not sameNsubgrp(out, new3):
            out = concat_tup(out,new3)
        else:
            F3=True
        #conclusion
        if F1 and F2 and F3:
            return out

#tested
def subgrp2dessin(H,xy):
    """
    H is a subgroup of a permutation group G generated by 
    two permutations from tuple xy; tuple xy represents 
    a group homomorphism from F_2 to G; the function returns 
    a tuple that represents a child's drawing corresponding 
    to the action of F_2 on the set G/H of left cosets of H in G;
    the index |G:H| is the degree of this child's drawing
    """
    G=PG(xy); x,y = xy 
    L = G.coset_transversal(H)
    aux_x = [L.index(G._coset_representative(a*x, H)) for a in L]
    aux_y = [L.index(G._coset_representative(a*y, H)) for a in L]
    return (permut(aux_x), permut(aux_y))

"""
Getting equivalence relations from the storage file..
"""
PB4_list=load_now('subGrPB4_org35') 
listE = [Equiv(tt) for tt in PB4_list]
print('     ')
print('listE is formed.')
print('The entries of listE are compatible equivalence relations on PaB(<=4) given in the object format.')
print('    ')
del PB4_list

"""
Getting the nested list of charming GT-shadows whose targets 
are elements of listE;
the i-th entry of GTcharm_wm is the list of charming 
GT-shadows with the target listE[i]; the 
GT-shadows are given in the tuple format 
"""

GTcharm_wm=load_now('wm_list_charm35')
print('The nested list GTcharm_wm of charming GT-shadows is uploaded.')
print('The GT-shadows in GTcharm_wm are given in the tuple format.')
print('   ')


"""
Getting the nested list of all GT-shadows for the 
first 31 elements of listE; the i-th entry of GTall_wm is the 
list of all GT-shadows with the target listE[i] for 0 <= i <= 30 in 
the tuple format.
"""

GTall_wm=load_now('wm_list_all31')
print('The nested list GTall_wm of all GT-shadows for the first 31 elements of listE is uploaded.')
print('The GT-shadows are given in the tuple format.')
print('    ')




print('We can now form a nested list GTcharm of all charming GT-shadows for the equivalence relations in listE.')
print('It may take more than 1 minute.')
ans = input('Please type yes if you want to form the nested list GTcharm: ')

if ans == 'yes':
    GTcharm = []
    t0=time.time()
    for i in range(35):
        GTcharm.append([GTsh(wm, listE[i].PB4) for wm in GTcharm_wm[i]])
    print('It took ', (time.time()-t0)/60, 'minutes')
    print('The nested list GTcharm is formed.')
    print('    ')
else:
    print('OK. I did NOT form the list GTcharm.')


#THE CODE BELOW WAS USED FOR TESTING:


"""
Testing via the virtual cyclotomic character...
"""
#tested
def test_cyclotomic(t):
    """
    the tuple t represents an element N of NFI_{PB_4}(B_4);
    the function returns true if the image of 
    the cyclotomic character from GT_{charm}(N) to the 
    group of units of the ring Z/N_{ord}Z is onto; 
    otherwise False
    """
    n0=Equiv(t).N0; tot=totient(n0)
    im=[] # this is for the values of the virtual cyclotomic character to im
    for wm in gener_GT_charm(t):
        val=(2*wm[1]+1)%n0
        if val not in im:
            im.append(val)
        if len(im)==tot:
            print(set(im)) #used for testing
            return True
    print('Something is wrong with the values of the virtual cyclotomic character.')
    return False

                
#used for testing
def test_conj_braid_rel(t):
    """
    t is a tuple of 6 permutations that represents a homomorphism 
    from PB_4 to S_d; the function returns True if all three relations
    sigma_1 sigma_2 sigma_1=..., sigma_2 sigma_3 sigma_2 =... and 
    sigma_1 sigma_3 = sigma_3 sigma_1 are satisfied for conjBySig1,
    conjBySig2, conjBySig3
    """
    R1=(conjBySig1(conjBySig2(conjBySig1(t)))==conjBySig2(conjBySig1(conjBySig2(t))))
    R2=(conjBySig2(conjBySig3(conjBySig2(t)))==conjBySig3(conjBySig2(conjBySig3(t))))
    R3= (conjBySig1(conjBySig3(t))==conjBySig3(conjBySig1(t)))
    return R1 and R2 and R3
        

#tested
def generPB4(d):
    """
    d is a positive integer;
    The function generates elements of a subset of homomorphisms
    from PB4 to S_d. It is quite close to a brute force way of 
    generating ALL homomorphisms from PB_4 to S_d. 
    It works very slowly but we can use it for testing.
    """
    Sd=SG(d)
    for t in split(ran(d)):
        x12=permut(t)
        for x23 in Sd.elements:
            for c3 in Sd.centralizer((x12,x23)).elements:
                x13 = compAll((x12**(-1),x23**(-1), c3))
                for x14, x24 in prod(SG(d).elements,repeat=2):
                    Z=Sd.centralizer((x12,x23,x13,x14,x24))
                    for c4 in Z.elements:
                        x34=compAll((c4,x24**(-1),x14**(-1),c3**(-1)))
                        tup = (x12,x23,x13,x14,x24,x34)
                        if relPB4(tup):
                            yield tup
                    
            

"""
#Here is how we tested generB4 using generB4slow:

#for n = 3
B4_3 = []

for t in generB4(3):
    for g in SG(3).elements:
        tt = conj_tup(g,t)
        if tt not in B4_3:
            B4_3.append(tt)
            

print(set(B4_3)==set(generB4slow(3))) # the output was True

#for n = 4
B4_4 = []

for t in generB4(4):
    for g in SG(4).elements:
        tt = conj_tup(g,t)
        if tt not in B4_4:
            B4_4.append(tt)
            
print(set(B4_4)==set(generB4slow(4))) # the output was True
#END of testing generB4
"""

"""
We used this for testing hexa1 and hexa2. 
The idea is that, if k=0 (i.e. f = f_xy(x12, x23) c^0) 
and the image of f belongs to [F_2/N_{F_2}, F_2/N_{F_2}] 
then the hexagon relations are equivalent to 

f(y,x) f(x,y)  \in N_{PB3}

f(z,x) z^m f(y,z) y^m f(x,y) x^m \in N_{PB_3}

"""
#used for testing
def hexa_test(wm, tt):
    """
    wm is a candidate for a shadow (with k = 0) and tt represents 
    an element of NFI_{PB_3}(B_3);the function returns True if the 
    GT-versions of hexagon relations are satisfied.
    """
    w,m = wm[0],wm[1]; x12, x23 = tt[:2]; xy = tt[:2]
    z = comp(x12,x23)**(-1)
    yx = (x23, x12); zx = (z,x12); yz = (x23,z)
    fxy = compAll(tuple(xy[i] for i in w))
    fyx = compAll(tuple(yx[i] for i in w))
    if not_id(fxy*fyx):
        return False
    fzx = compAll(tuple(zx[i] for i in w))
    fyz = compAll(tuple(yz[i] for i in w))
    tup = (fzx, z**m, fyz, x23**m, fxy, x12**m)
    return compAll(tup).order()==1
    

"""
More testing of the functions hexa1 and hexa2
Details? 
"""


#used for testing
def hhexa1(wm,tt):
    """
    wm is a candidate for a GT-pair and tt represents an element of NFI_{PB_3}(B_3)
    The function returns True if the first hexagon is satisfied.
    Otherwise, False.
    """
    if wm == ((),0):
        # ((),0) corresponds to the identity map PaB--> PaB
        return True
    c=cenPB3(tt)
    xy = tt[:2]; x12, x23 = xy
    z = comp(x12, x23)**(-1); w,m = wm
    xzc= (x12,z*c); yzc = (x23,z*c) 
    fxy = compAll(tuple(xy[i] for i in w)) 
    fxzc_inv = compAll(tuple(xzc[i] for i in w))**(-1)
    fyzc= compAll(tuple(yzc[i] for i in w))
    tup = (x23**m, fxy, x12**m,fxzc_inv, z**m, fyzc)
    return compAll(tup).order()==1 # only the identity elem has order 1


#used for testing
def hhexa2(wm, tt):
    """
    wm is a candidate for a GT-pair and tt represents an element of NFI_{PB_3}(B_3)
    The function returns True if the second hexagon is satisfied. 
    Otherwise, False.
    """
    if wm == ((),0):
        # ((),0) corresponds to the identity map PaB--> PaB
        return True
    c=cenPB3(tt)
    xy = tt[:2]; x12, x23 = xy
    u = comp(x23, x12)**(-1); w,m = wm
    ucx = (u*c,x12); ucy = (u*c,x23)
    fxy_inv = compAll(tuple(xy[i] for i in w))**(-1)
    fucx_inv = compAll(tuple(ucx[i] for i in w))**(-1)
    fucy = compAll(tuple(ucy[i] for i in w))
    tup = (fucx_inv, x12**m, fxy_inv, x23**m, fucy, u**m)
    return compAll(tup).order()==1
    


"""
Testing generWF2 and generWComm
"""
#used for testing the generator generWF2
def test_generWF2(d):
    t = (prm(d), prm(d)); G = PG(t)
    sz = G.order()
    print('The size of the group is ',sz)
    print('    ')
    print('I will now form the list of words...')
    Wlist=list(generWF2(t,True)) # list of words
    print('    ')
    if sz==len(Wlist):
        print('The total number of words is correct.')
    Eset={w2g(w,t) for w in Wlist}
    if Eset==G.elements:
        print('We got the correct set of elements.')
    
    
#used for testing the generator generWF2
def test_generWComm(d):
    t = (prm(d), prm(d)); G = PG(t)
    H=G.commutator(G,G)
    sz = H.order()
    print('The size of the commutator subgroup is ',sz)
    print('    ')
    print('I will now form the list of words...')
    Wlist=list(generWComm(t,True)) # list of words
    print('    ')
    if sz==len(Wlist):
        print('The total number of words is correct.')
    Eset={w2g(w,t) for w in Wlist}
    if Eset==H.elements:
        print('We got the correct set of elements.')

"""
the function below was used for testing various generators 
of child's drawings...
"""

#use for testing
def test_dessin(d):
    """
    the command will work for a long time if d > 6
    """
    Tup=all_dessin(d)
    Tup1=all_dessin_slow(d)
    DL=[]; DL1=[]
    for c in Tup:
        D = Dessin(c)
        if D not in DL:
            DL.append(D)
    for c in Tup1:
        D=Dessin(c)
        if D not in DL1:
            DL1.append(D)
    Test={D in DL for D in DL1}
    print(len(DL)==len(DL1) and {True}==Test)


"""
THE CODE BELOW IS NO LONGER USED:...
"""

